module Make
  (Lattice : Lattice.S)
= struct

(* Repeat the type definitions given in the interface. *)

module Point =
  Label

type point =
    Point.t

type point_set =
    Point.Set.t

type 'a point_map =
    'a Point.Map.t

type value =
    Lattice.t

type valuation =
    point -> value

type map =
    value point_map

module type GeneralProblem = sig
  val transfer: valuation -> valuation
  val successors: point -> point_set
  val entry: point_set
end

module type SimplifiedProblem = sig
  val successors: point -> point_set
  val predecessors: point -> point_set
  val semantics: point -> value -> value
  val seed: point -> value
  val entry: point_set
end

(* ------------------------------------------------------------------------- *)
(* Intensional and extensional views of valuations. *)

(* [m2v] converts a map [m] into a valuation. Points that do not
   explicitly appear in the domain of [m] are mapped to the lattice's
   bottom element. *)

let m2v (m : map) : valuation =
  function p ->
    try
      Point.Map.find p m
    with Not_found ->
      Lattice.bottom

(* ------------------------------------------------------------------------- *)
(* Here is the solver for general problems. *)

module General (P : GeneralProblem) = struct

  (* We maintain a waiting pool, a set of points that should be
     examined or re-examined.

     An invariant is that if a point [p] is not a descendant of a
     point in the pool, then the current valuation at [p] has reached
     its final value -- it is the solution at [p]. Initially, the pool
     is [P.entry], so no point [p] in the domain satisfies the
     invariant's precondition, and the invariant is true. When the
     pool becomes empty, every point [p] satisfies the invariant's
     precondition, so the current valuation is the solution.

     Because the waiting pool is a set, no point appears twice in it
     (which would be a source of inefficiency).

     For simplicity, points are taken out of the waiting pool in an
     arbitrary order. A more elaborate approach would consist in
     dealing with high-priority points first, where a point receives
     higher priority than its descendants in the graph (unless cycles
     in the graph make that impossible). Such priorities could be
     assigned by performing a depth-first traversal of the dependency
     graph. *)

  (* Here is the iterative algorithm. As long as the pool is nonempty,
     we take a point [p] off the pool and examine it. The variable [m]
     holds the current valuation. Its initial value is the empty map,
     which is interpreted as bottom everywhere. *)

  let solution : valuation =
    m2v (
      Point.Set.exhaust P.entry Point.Map.empty (fun p (m : map) ->

	(* Evaluate the image of the current valuation through the
	   transfer function at [p]. Compute its join with the value of
	   the current valuation at [p], and check whether this new
	   value is strictly greater than the old value. *)

	let valuation = m2v m in
	let old_value = valuation p in
	let new_value = Lattice.join old_value (P.transfer valuation p) in

	(* If the two values coincide, there is nothing to do about
	   [p], as it is in agreement with its predecessors. If they
	   do not coincide, then we must update the current valuation
	   at [p] and schedule [p]'s successors for examination or
	   re-examination. *)

	if Lattice.equal old_value new_value then

	  (* Continue without adding any points to the waiting pool
	     and with the current map. *)

	  Point.Set.empty, m

	else

	  (* Schedule [p]'s successors for examination and Continue
	     with an updated map. *)

	  P.successors p, Point.Map.add p new_value m

      )
    )

end

(* ------------------------------------------------------------------------- *)
(* Here is the solver for simplified problems. *)

module Simplified (P : SimplifiedProblem) = struct

  (* Let us transform the simplified problem into a general problem.

     We choose to solve for the [in] valuation. One could also solve
     for the [out] valuation, or solve for the pointwise product of
     both. These choices do not make much difference, since either
     valuation can be defined in terms of the other. *)

  (* This transforms an [in] valuation into the corresponding [out]
     valuation. This is done simply by applying the semantics
     function. *)

  let in2out (invaluation : valuation) : valuation =
    function p ->
      P.semantics p (invaluation p)

  module GP = struct

    (* Define the transfer function. To obtain the new value of the
       [in] valuation at [p], we compute the join of the seed at [p]
       and of the join, over all predecessors [q] of [p], of the
       previous value of the [out] valuation at [q] -- which itself is
       defined in terms of the previous value of the [in] valuation at
       [q]. *)

    let transfer (invaluation : valuation) : valuation =
      let outvaluation = in2out invaluation in
      function p ->
	Point.Set.fold (fun q value ->
	  Lattice.join value (outvaluation q)
        ) (P.predecessors p) (P.seed p)

    (* Transmit information about successors and domain. Note that the
       successors of [p] in a syntactic sense -- the endpoints of
       [p]'s outgoing edges -- are indeed the successors of [p] in a
       semantic sense -- they are the points where evaluating the
       image of a valuation through the above transfer function
       requires evaluating that valuation at [p]. *)

    let successors = P.successors
    let entry = P.entry

  end

  (* Solve the general problem. *)

  module S = General(GP)

  (* Transmit the final [in] valuation, and compute the final [out]
     valuation in terms of it. *)

  let insolution =
    S.solution

  let outsolution =
    in2out insolution

end

(* ------------------------------------------------------------------------- *)

end

