(* This module provides algorithms that compute solutions of data flow
   equations, that is, fixpoints of abstract transfer functions. *)

(* We are interested in computing so-called valuations, which are
   functions of so-called points (represented as control flow graph
   labels) into a lattice of so-called abstract values. The entire
   module is parameterized over the lattice, so as to enable analyses
   that require different lattices.  *)

module Make
  (Lattice : Lattice.S)
: sig

  (* Points are control flow graph labels. *)

  type point =
      Label.t

  type point_set =
      Label.Set.t

  type 'a point_map =
      'a Label.Map.t

  (* Abstract values are lattice elements. *)

  type value =
      Lattice.t

  (* A valuation is a mapping of a fixed domain (a set of points)
     into lattice elements. The lattice ordering is extended
     pointwise to valuations. *)

  (* The type [valuation] specifies an intensional representation of
     valuations, that is, a view of valuations as functions. *)

  type valuation =
      point -> value

  (* The type [map] specifies an extensional representation of
     valuations, that is, a view of valuations as data structures. *)

  type map =
      value point_map

  (* A so-called general analysis problem is defined as follows. *)

  module type GeneralProblem = sig

    (* A transfer function is a monotonic valuation transformer. *)

    val transfer: valuation -> valuation

    (* When the image of a valuation through the transfer function is
       evaluated at [p], the valuation itself is evaluated only at a
       finite number of points, which form a possibly strict subset of
       the domain. We refer to these points as the predecessors of
       [p].

       Conversely, when a valuation is modified at [p], its image
       through the transfer function is affected only at the
       successors of [p].

       In summary, predecessor (or successor) edges reflect
       dependencies in the transfer function. These dependencies are
       made explicit by the [successors] function below.

       As far as correctness is concerned, it is fine to only offer a
       conservative approximation of the predecessor (or successor)
       sets. For instance, taking the predecessor (or successor) set
       of every point to be the full domain would be a correct
       approximation. However, the analysis runs more efficiently when
       a more precise approximation is provided. *)

    val successors: point -> point_set

    (* The problem domain is the set of points of interest. It is the
       least set that contains [entry] and is closed under successor
       edges. *)

    val entry: point_set

  end

  (* This module defines a solver for general analysis problems. *)

  module General (P : GeneralProblem) : sig

    (* The (unique) solution to the above general analysis problem is
       the least valuation that is greater than (or equal to) its own
       image through the transfer function.

       If queried about a point that does not belong to the problem
       domain -- for instance, a point that is unreachable from any of
       the entry points -- the solution returns bottom. *)

    val solution: valuation

  end

  (* In a general analysis problem, the transfer function is the
     central component of the problem definition, and a dependency
     graph arises out of the transfer function. In a simplified
     analysis problem, things are somewhat more concrete: we start
     with a graph (typically a control flow graph), associate a
     semantics with every point (typically an abstraction of the
     semantics of the instruction found at that point), associate a
     seed with every point, and let a transfer function arise out of
     these definitions.

     The simplified problem formulation is somewhat less general than
     the original one, which is why we expose both in this interface.

     In particular, in a simplified analysis problem, the information
     that is transmitted by a node to its successors is the same along
     all successor edges, whereas, in a general analysis problem, each
     successor can be sent different abstract values: for instance, an
     [if] instruction could send information that reflects the success
     or the failure of the test to each of its two successors. *)

  (* A so-called simplified analysis problem is defined as follows. *)

  module type SimplifiedProblem = sig

    (* Let points form a graph. Points typically are control flow
       graph labels, and edges typically represent control flow graph
       edges. *)

    val successors: point -> point_set
    val predecessors: point -> point_set

    (* Let every point carry a monotonic semantics function that maps
       the lattice into itself. The semantics function associated with
       a point [p] typically reflects the semantics of the instruction
       found at [p] in the control flow graph. It transforms
       information that holds before the instruction is executed into
       information that holds after the instruction is executed. *)

    val semantics: point -> value -> value

    (* Let every point carry a lattice element, the seed. *)

    val seed: point -> value

    (* The above information allows defining a transfer function
       over pairs of an [in] valuation and an [out] valuation, or,
       equivalently, over a valuation that maps every point to a
       pair of an [in] value and an [out] value. The [in] value
       and the [out] value are related as follows: at every point
       [p], the [out] value is the image of the [in] value through
       the semantics function.

       Now, the transfer function is defined, in slightly informal
       speak, as follows: the new [in] value at [p] is the join of the
       seed at [p] and of the join, over all predecessors [q] of [p],
       of the old [out] value at [q].

       This definition explains how a simplified analysis problem
       is viewed as a general analysis problem. *)

    (* The problem domain is defined like above. *)

    val entry: point_set

  end

  (* This module defines a solver for simplified analysis problems. *)

  module Simplified (P : SimplifiedProblem) : sig

    (* These are the [in] and [out] solutions to the above simplified
       analysis problem.

       If queried about a point that does not belong to the problem
       domain -- for instance, a point that is unreachable from any of
       the entry points -- the solutions return bottom. *)

    val insolution: valuation
    val outsolution: valuation

  end

end

