open ERTL

(* These functions allow turning an [ERTL] control flow graph into an
   explicit graph, that is, making successor edges explicit. This is
   useful in itself and facilitates the computation of predecessor
   edges. *)

let instruction_successors (i : instruction) =
  match i with
  | IReturn _ ->
      Label.Set.empty
  | INewFrame l
  | IDeleteFrame l
  | IGetHwReg (_, _, l)
  | ISetHwReg (_, _, l)
  | IGetStack (_, _, l)
  | ISetStack (_, _, l)
  | IGetGlobal (_, _, l)
  | ISetGlobal (_, _, l)
  | IConst (_, _, l)
  | IUnOp (_, _, _, l)
  | IBinOp (_, _, _, _, l)
  | ICall (_, _, l)
  | ILoad (_, _, _, l)
  | IStore (_, _, _, l)
  | IGoto l ->
      Label.Set.singleton l
  | IUnBranch (_, _, l1, l2)
  | IBinBranch (_, _, _, l1, l2) ->
      Label.Set.couple l1 l2

let successors (graph : graph) : Label.SetMap.t =
  Label.Map.map instruction_successors graph

(* The analysis uses the lattice of pairs of a subset of the
   function's pseudo-registers and a subset of the hardware
   registers. The lattice's join operation is pointwise set union,
   which reflects the fact that a pseudo-register or hardware register
   is deemed live at a program point if and only if it is live at any
   of the successors of that program point. *)

module L = struct

  type t =
      Register.Set.t * MIPS.RegisterSet.t

  let bottom =
    Register.Set.empty, MIPS.RegisterSet.empty

  let psingleton r =
    Register.Set.singleton r, MIPS.RegisterSet.empty

  let hsingleton hwr =
    Register.Set.empty, MIPS.RegisterSet.singleton hwr

  let join (rset1, hwrset1) (rset2, hwrset2) =
    (Register.Set.union rset1 rset2, MIPS.RegisterSet.union hwrset1 hwrset2)

  let diff (rset1, hwrset1) (rset2, hwrset2) =
    (Register.Set.diff rset1 rset2, MIPS.RegisterSet.diff hwrset1 hwrset2)

  let equal (rset1, hwrset1) (rset2, hwrset2) =
    Register.Set.equal rset1 rset2 && MIPS.RegisterSet.equal hwrset1 hwrset2

end

(* These are the sets of pseudo-registers and hardware registers
   defined at (written by) an instruction. Note that the [ICall]
   instruction potentially destroys all caller-save hardware
   registers. *)

let defined (i : instruction) : L.t =
  match i with
  | IGetHwReg (r, _, _)
  | IGetStack (r, _, _)
  | IConst (r, _, _)
  | IUnOp (_, r, _, _)
  | IBinOp (_, r, _, _, _)
  | ILoad (r, _, _, _)
  | IGetGlobal (r, _, _) ->
      L.psingleton r
  | ICall _ ->
      Register.Set.empty, MIPS.caller_saved
  | ISetHwReg (hwr, _, _) ->
      L.hsingleton hwr
  | INewFrame _
  | IDeleteFrame _
  | ISetStack _
  | IStore _
  | ISetGlobal _
  | IGoto _
  | IUnBranch _
  | IBinBranch _
  | IReturn _ ->
      L.bottom

(* This is the set of pseudo-registers and hardware registers used at
   (read by) an instruction. Note that the [ICall] instruction reads
   the hardware registers that are used to pass actual
   parameters. [IReturn flag] reads not only [$ra], but also the
   callee-save registers, and the result-value register [$v0] (if
   [flag] is set). This is because the caller expects to find
   meaningful values in these registers, so they must be considered
   live immediately before the [IReturn] instruction. *)

let used (i : instruction) : L.t =
  match i with
  | IGetHwReg (_, hwr, _) ->
      L.hsingleton hwr
  | INewFrame _
  | IDeleteFrame _
  | IGetStack _
  | IConst _
  | IGoto _
  | IGetGlobal _ ->
      L.bottom
  | IReturn with_a_result ->
      Register.Set.empty,
      MIPS.RegisterSet.add_if with_a_result MIPS.result (
	MIPS.RegisterSet.add MIPS.ra (
	  MIPS.callee_saved
        )
      )
  | ISetHwReg (_, r, _)
  | ISetStack (_, r, _)
  | IUnOp (_, _, r, _)
  | ILoad (_, r, _, _)
  | ISetGlobal (_, r, _)
  | IUnBranch (_, r, _, _) ->
      L.psingleton r
  | IBinOp (_, _, r1, r2, _)
  | IStore (r1, _, r2, _)
  | IBinBranch (_, r1, r2, _, _) ->
      Register.Set.couple r1 r2, MIPS.RegisterSet.empty
  | ICall (_, nparams, _) ->
      Register.Set.empty, MIPS.RegisterSet.of_list (Misc.prefix nparams MIPS.parameters)

(* An instruction is considered pure if it has no side effect, that is, if
   its only effect is to write a value to its destination pseudo-register or
   hardware register.

   A pure instruction whose destination is dead after the instruction will
   be eliminated during the translation of [ERTL] to [LTL]. This is done by
   replacing the instruction with an [IGoto] instruction.

   [eliminable liveafter i] returns [Some l], where [l] is [i]'s single
   successor, if instruction [i] is eliminable. Otherwise, it returns
   [None]. The parameter [liveafter] is the set of variables that are live
   after the instruction. *)

let eliminable ((pliveafter, hliveafter) : L.t) (i : instruction) =
  match i with
  | IGetHwReg (r, _, l)
  | IGetStack (r, _, l)
  | IGetGlobal (r, _, l)
  | IConst (r, _, l)
  | IUnOp (_, r, _, l)
  | IBinOp (_, r, _, _, l)
  | ILoad (r, _, _, l) ->
      if Register.Set.mem r pliveafter then None else Some l
  | ISetHwReg (hwr, _, l) ->
      if MIPS.RegisterSet.mem hwr hliveafter then None else Some l
  | IReturn _
  | INewFrame _
  | IDeleteFrame _
  | ISetStack _
  | ISetGlobal _
  | ICall _
  | IStore _
  | IGoto _
  | IUnBranch _
  | IBinBranch _ ->
      None

(* This is the abstract semantics of instructions. It defines the
   [out] value of an instruction in terms of its [in] value. That is,
   since liveness analysis is a backward analysis, it defines the
   pseudo-registers that are live before the instruction in terms of
   those that are live after the instruction.

   The standard definition is: a pseudo-register is considered live
   before the instruction if either (1) it is used by the instruction,
   or (2) it is live after the instruction and not defined by the
   instruction.

   As an exception to this rule, if the instruction is eliminable,
   then a pseudo-register is considered live before the instruction
   if and only if it is live after the instruction. This anticipates
   on the instruction's elimination.

   This exception means that the source pseudo-registers of a pure
   instruction need not be considered live if the instruction's result
   is unused. This allows a sequence of pure instructions whose end
   result is dead to be considered entirely dead.

   It is a simple, but not entirely trivial, exercise to check that
   this transfer function is monotone. *)

let instruction_semantics (i : instruction) (invalue : L.t) : L.t =
  match eliminable invalue i with
  | None ->
      L.join (L.diff invalue (defined i)) (used i)
  | Some _ ->
      invalue

(* Let us specialize the data flow analysis problem solver for this
   lattice. *)

module K =
  Kildall.Make (L)

type valuation =
    K.valuation

(* This is how we turn an [ERTL] procedure into a liveness analysis
   problem and solve it. *)

let analyze (proc : procedure) : valuation * valuation =

  (* Turn the procedure's control flow graph into explicit successor
     and predecessor graphs. *)

  let successors = successors proc.graph in
  let predecessors = Label.SetMap.reverse successors in

  (* Formulate and solve the analysis problem. *)

  let module S = K.Simplified (struct

    (* Since this is a backward analysis, an instruction's successors
       in the analysis problem are its predecessors in the control
       flow graph, and vice versa. *)

    let successors label =
      Label.SetMap.find label predecessors

    and predecessors label =
      Label.SetMap.find label successors

    let semantics label =
      instruction_semantics (Label.Map.find label proc.graph)

    (* The seed is bottom everywhere. *)

    let seed label =
      L.bottom

    (* The problem domain is the entire control flow graph. We do not
       assume that all points are reachable from the exit point --
       that would not be true in a procedure with an infinite loop,
       for instance. *)

    let entry =
      Label.Map.domain proc.graph

  end) in

  (* Transmit the solutions. Because this is a backward analysis, the
     [in] solution reported by Kildall's algorithm is our [after]
     solution, and the [out] solution reported by Kildall's algorithm
     is our [before] solution. *)

  S.outsolution, S.insolution

