(* This module performs liveness analysis over the control flow
   graph of a single [ERTL] procedure.

   We collect liveness information about the procedure's
   pseudo-registers and about allocatable hardware registers. We are
   not interested in collecting information about non-allocatable
   hardware registers such as [$gp], [$sp], etc. so they are
   considered never defined and never used as far as [ERTL] is
   concerned. *)

open ERTL

(* This is the lattice of pairs of a subset of the function's
   pseudo-registers and a subset of the hardware registers. *)

module L : sig

  include Lattice.S with type t = Register.Set.t * MIPS.RegisterSet.t

  val diff: t -> t -> t
  val psingleton: Register.t -> t
  val hsingleton: MIPS.register -> t

end

(* [defined i] is the set of pseudo-registers and hardware registers
   defined at (written by) instruction [i]. *)

val defined: instruction -> L.t

(* A valuation is a function that maps a program point (a control flow
   graph label) to the set of pseudo-registers and hardware registers
   that are live at that point. More precisely, a [before] valuation
   tells which registers are live immediately before the instruction
   at that point is executed, while an [after] valuation tells which
   registers are live immediately after the instruction at that point
   is executed. *)

type valuation =
    Label.t -> L.t

(* [analyze proc] analyzes the procedure [proc] and returns a pair of
   a [before] valuation and an [after] valuation. *)

val analyze: procedure -> valuation * valuation

(* Pure instructions whose destination pseudo-register is dead after the
   instruction will be eliminated during the translation of [ERTL] to [LTL].
   [eliminable liveafter i] returns [Some successor], where [successor] is
   [i]'s single successor, if instruction [i] is eliminable. Otherwise, it
   returns [None]. *)

val eliminable: L.t -> instruction -> Label.t option

