open MIPSOps

(* ------------------------------------------------------------------------- *)

(* [translate_instruction] translates an [LTL] instruction into a
   [LIN] instruction. *)

let translate_instruction = function

    (* Sequential instructions. *)

  | LTL.INewFrame _ ->
      LIN.INewFrame
  | LTL.IDeleteFrame _ ->
      LIN.IDeleteFrame
  | LTL.IGetStack (r, slot, _) ->
      LIN.IGetStack (r, slot)
  | LTL.ISetStack (slot, r, _) ->
      LIN.ISetStack (slot, r)
  | LTL.IConst (r, i, _) ->
      LIN.IConst (r, i)
  | LTL.IUnOp (op, r1, r2, _) ->
      LIN.IUnOp (op, r1, r2)
  | LTL.IBinOp (op, r, r1, r2, _) ->
      LIN.IBinOp (op, r, r1, r2)
  | LTL.ICall (callee, _) ->
      LIN.ICall callee
  | LTL.ILoad (r1, r2, o, _) ->
      LIN.ILoad (r1, r2, o)
  | LTL.IStore (r1, o, r2, _) ->
      LIN.IStore (r1, o, r2)

    (* Because [Branch.compress] has been called before, no [IGoto]
       instruction can be reached. *)

  | LTL.IGoto _ ->
      assert false

    (* Conditional branch instructions. In [LIN], control implicitly
       falls through to the second successor, so only the first
       successor is explicitly mentioned in the instruction. *)

  | LTL.IUnBranch (cond, r, l1, _) ->
      LIN.IUnBranch (cond, r, l1)
  | LTL.IBinBranch (cond, r1, r2, l1, _) ->
      LIN.IBinBranch (cond, r1, r2, l1)

    (* Instructions without a successor. *)

  | LTL.IReturn ->
      LIN.IReturn

(* ------------------------------------------------------------------------- *)

(* The functor [Visit] implements the core of the translation. *)

module Visit (S : sig

  (* [require l] records the fact that the label [l] should explicitly
     exist in the [LIN] program. It must be used whenever a [LIN]
     branch instruction is issued. *)

  val require: Label.t -> unit

  (* [mark l] marks the label [l]. [marked l] tells whether [l] is
     marked. *)

  val mark: Label.t -> unit
  val marked: Label.t -> bool

  (* [fetch l] is the instruction found at label [l] is in the source
     program. *)

  val fetch: Label.t -> LTL.instruction

  (* [generate i] generates instruction [i]. Instructions are
     generated sequentially. *)

  val generate: LIN.instruction -> unit

end) : sig

  (* [visit] implements a depth-first traversal of the control flow graph,
     generating instructions as new nodes are being discovered.

     If label [l] has already been discovered, then [visit l] issues
     an [IGoto] instruction towards [l]. If [l] has not been
     discovered yet, [visit l] marks [l] as discovered, issues an
     [ILabel] instruction, translates the instruction found at [l] in
     the source program, and visits its successors. *)

  val visit: Label.t -> unit

end = struct

  open S

  let rec visit l =

    if marked l then begin

      (* Label [l] has been visited before. This implies that an [ILabel l]
	 instruction has been issued already. We must now generate an
	 [IGoto] instruction that transfers control to this place. Because
	 [l] is the target of a branch instruction, we require it to exist
	 in the [LIN] code. *)

      require l;
      generate (LIN.IGoto l)

    end
    else begin

      (* Label [l] has never been visited before. First, record that it
	 now has been visited, so as to avoid looping. *)

      mark l;

      (* Then, generate an [ILabel l] instruction. This instruction
	 will be useless if [l] turns out not to be the target of a
	 branch: this is taken care of later. *)

      generate (LIN.ILabel l);

      (* Fetch the instruction found at label [l] in the source program. *)

      let instruction = fetch l in

      (* As an optional optimization, check if this is a conditional branch
	 whose implicit successor has been visited before and whose explicit
	 successor has not. In that case, if we did nothing special, we would
	 produce a conditional branch immediately followed with an
	 unconditional one, like this:

		bgtz  $t1, find24
		j     find42
		find24:
                ...

	 This can be avoided simply by reversing the condition:

                blez  $t1, find42
                ...

	 *)

      let instruction =
	match instruction with

	| LTL.IUnBranch (cond, r, l1, l2)
	  when marked l2 && not (marked l1) ->
	    let cond =
	      match cond with
	      | UConGez ->
		  UConLtz
	      | UConGtz ->
		  UConLez
	      | UConLez ->
		  UConGtz
	      | UConLtz ->
		  UConGez
	    in
	    LTL.IUnBranch (cond, r, l2, l1)

	| LTL.IBinBranch (cond, r1, r2, l1, l2)
	  when marked l2 && not (marked l1) ->
	    let cond =
	      match cond with
	      | ConEq ->
		  ConNe
	      | ConNe ->
		  ConEq
	    in
	    LTL.IBinBranch (cond, r1, r2, l2, l1)

	| _ ->
	    instruction
      in

      (* Translate the instruction. *)

      generate (translate_instruction instruction);

      (* Note that [translate_instruction] never produces an [IGoto]
	 instruction. As a result, the code above never generates an [ILabel]
	 instruction immediately followed with an [IGoto] instruction. This
	 proves that we never generate a (conditional or unconditional) branch
	 towards an [IGoto] instruction. *)

      (* There now remains to visit the instruction's successors. *)

      match instruction with

	(* Sequential instructions. There is only one successor, with implicit
	   fallthrough. *)

      | LTL.INewFrame l
      | LTL.IDeleteFrame l
      | LTL.IGetStack (_, _, l)
      | LTL.ISetStack (_, _, l)
      | LTL.IConst (_, _, l)
      | LTL.IUnOp (_, _, _, l)
      | LTL.IBinOp (_, _, _, _, l)
      | LTL.ICall (_, l)
      | LTL.ILoad (_, _, _, l)
      | LTL.IStore (_, _, _, l)
      | LTL.IGoto l ->

	  visit l

	(* Conditional branch instructions. The label that is reached by
	   falling through in [LIN] is [l2], which means that it must be
	   translated first, so that its instructions are contiguous with the
	   [LIN] branch instruction. The label [l1], on the other hand,
	   becomes the target of a jump, so it is required to exist in the
	   [LIN] code.

	   Code for [l1] is generated, if necessary, after we are done dealing
	   with [l2]. If [l1] has already been visited at this point, no code
	   needs be produced, so the second call to visit is made only if [l1]
	   has not been visited yet. *)

      | LTL.IUnBranch (_, _, l1, l2)
      | LTL.IBinBranch (_, _, _, l1, l2) ->

	  visit l2;
	  require l1;
	  if not (marked l1) then
	    visit l1

	(* Instructions without a successor. *)

      | LTL.IReturn ->

	  ()

    end

end

(* ------------------------------------------------------------------------- *)

(* [translate entry graph] turns an [LTL] control flow graph into
   a [LIN] sequence of instructions. *)

let translate entry graph =

  (* Keep track of the labels that have been visited (initially none), of the
     labels that must exist within the [LIN] code (initially only the graph's
     entry point) and of the list of [LIN] instructions that are being
     generated (initially empty). Instructions are held in the list in reverse
     order, for efficiency. The list is reversed once generation is over. *)

  let visited, required, instructions =
    ref Label.Set.empty, ref (Label.Set.singleton entry), ref []
  in

  (* Instantiate the functor. *)

  let module V = Visit (struct
    let require label =
      required := Label.Set.add label !required
    let mark label =
      visited := Label.Set.add label !visited
    let marked label =
      Label.Set.mem label !visited
    let fetch label =
      Label.Map.find label graph
    let generate instruction =
      instructions := instruction :: !instructions
  end) in

  (* Traverse the control flow graph. *)

  V.visit entry;

  (* Now, get rid of the labels that do not need to exist. Also,
     reverse the list to re-establish the correct order. *)

  List.filter (function
    | LIN.ILabel l ->
	Label.Set.mem l !required
    | _ ->
	true
  ) (List.rev !instructions)

(* ------------------------------------------------------------------------- *)

(* Extend the translation to procedures and programs. *)

let translate_procedure proc =
  let entry, graph = Branch.compress proc.LTL.entry proc.LTL.graph in
  {
    LIN.formals = proc.LTL.formals;
    LIN.locals = proc.LTL.locals;
    LIN.code = translate entry graph
  }

let translate_program (p : LTL.program) : LIN.program =
  {
    LIN.globals = p.LTL.globals;
    LIN.defs = StringMap.map translate_procedure p.LTL.defs
  }

